/* Orion Lawlor's Short UNIX Examples, olawlor@acm.org 2004/2/16

Shows how to use POSIX siginfo_t to find the faulting
address from a signal handler.  Also prints out all the
other machine registers.
*/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <ucontext.h>
#ifdef SOLARIS /* needed with at least Solaris 8 */
#include <siginfo.h>
#endif

const static struct {
	int sigNo;
	const char *name;
} sigs[]={
	{SIGSEGV,"SIGSEGV-- segmentation violation (bad memory access)"},
	{SIGBUS,"SIGBUS-- bus error (misaligned memory access)"},
	{SIGILL,"SIGILL-- illegal instruction (jump to bad memory)"},
	{SIGFPE,"SIGFPE-- floating-point exception (divide by zero)"}
};

const char *sig2name(int sig) {
	int i,n=sizeof(sigs)/sizeof(sigs[0]); /* signal table length */
	for (i=0;i<n;i++) if (sig==sigs[i].sigNo) return sigs[i].name;
	return "Unknown signal";
}

void handler(int sig, siginfo_t *HowCome, void *ucontextv) {
	struct ucontext *uc=(struct ucontext *)ucontextv;
	struct sigcontext *mc=(struct sigcontext *)(&uc->uc_mcontext);
	printf("-------------------\n");
	printf("Caught signal signal %s (#%d)\n",sig2name(sig),sig);
	printf("Registers:  eip=0x%08x\n"
		"  eax=0x%08x  ebx=0x%08x  ecx=0x%08x   edx=0x%08x\n"
		"  esp=0x%08x  ebp=0x%08x  esi=0x%08x   edi=0x%08x\n"
		"  eflags=0x%08x\n",
		mc->eip,
		mc->eax,mc->ebx,mc->ecx,mc->edx,
		mc->esp_at_signal,mc->ebp,mc->esi,mc->edi,
		mc->eflags);
	
	printf("Signal raised at address %p.  Exiting.\n", HowCome->si_addr);
	exit(99);
}

void handle_signals(void) {
	int i,n=sizeof(sigs)/sizeof(sigs[0]); /* signal table length */
	stack_t stk;
	stk.ss_size=2*SIGSTKSZ;
	stk.ss_sp=malloc(stk.ss_size);
	stk.ss_flags=0;
	sigaltstack(&stk,0);
	for (i=0;i<n;i++) {
        	struct sigaction sa;
        	sa.sa_sigaction = handler;
        	sigemptyset( &sa.sa_mask );
        	sa.sa_flags = 0;
		sa.sa_flags |= SA_SIGINFO; /* we want a siginfo_t */
		sa.sa_flags |= SA_RESTART; /* restart interrupted syscalls (no EINTR errors) */
		sa.sa_flags |= SA_ONSTACK; /* use alternate signal stack */
        	if (sigaction (sigs[i].sigNo, &sa, 0)) {
			perror("sigaction");
			exit(1);
        	}
	}
}

int main(){ 
	int *where=(int *)0x123;
	
/* Install our signal handler */
	handle_signals();

/* Do something that generates a segfault */
        printf( "About to write to %p\n", where);
        *where = 'a';

        return(0);
}

/*<@>
<@> ******** Program output: ********
<@> About to write to 0x123
<@> -------------------
<@> Caught signal signal SIGSEGV-- segmentation violation (bad memory access) (#11)
<@> Registers:  eip=0x08048748
<@>   eax=0x00000018  ebx=0x08049d18  ecx=0x00000000   edx=0x00000018
<@>   esp=0xbf9a0110  ebp=0xbf9a0128  esi=0xbf9a01c4   edi=0xbf9a0150
<@>   eflags=0x00210292
<@> Signal raised at address 0x123.  Exiting.
<@> */
